/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2007 William Jon McCann <mccann@jhu.edu>
 * Copyright (C) 2010 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include "config.h"

#include <sys/types.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include <locale.h>

#include <glib.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <X11/Xatom.h>
#include <X11/extensions/XTest.h>
#include <X11/keysym.h>
#include <xorg/Xwacom.h>

#include "gnome-settings-profile.h"
#include "gsd-wacom-manager.h"
#include "gsd-input-helper.h"
#include "gsd-keygrab.h"
#include "gsd-wacom-device.h"
#include "gsd-wacom-devmanager-helpers.h"
#include "gsd-wacom-gdk-helpers.h"

#define GSD_WACOM_MANAGER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GSD_TYPE_WACOM_MANAGER, GsdWacomManagerPrivate))

#define GCONF_WACOM_DIR			"/desktop/gnome/peripherals/wacom"
#define GCONF_WACOM_DIR_SLASH		GCONF_WACOM_DIR "/"
#define WACOM_DEVICE_CONFIG_BASE	GCONF_WACOM_DIR_SLASH "%s/"

#define KEY_ROTATION			"rotation"
#define KEY_TOUCH			"touch"
#define KEY_TPCBUTTON			"tablet-pc-button"
#define KEY_IS_ABSOLUTE			"is-absolute"
#define KEY_AREA			"area"
#define KEY_PAD_BUTTON_MAPPING		"pad-buttonmapping"
#define KEY_DISPLAY			"display"

/* Stylus and Eraser settings */
#define KEY_BUTTON_MAPPING		"buttonmapping"
#define KEY_PRESSURETHRESHOLD		"pressurethreshold"
#define KEY_PRESSURECURVE		"pressurecurve"

/* Button settings */
#define KEY_ACTION_TYPE			"action-type"
#define KEY_CUSTOM_ACTION		"custom-action"
#define KEY_CUSTOM_ELEVATOR_ACTION	"custom-elevator-action"

/* See "Wacom Pressure Threshold" */
#define DEFAULT_PRESSURE_THRESHOLD 	27

struct GsdWacomManagerPrivate
{
	guint start_idle_id;
	GsdDeviceManager *device_manager;
	guint device_added_id;
	guint device_removed_id;
	guint notify_wacom;
	GHashTable *devices; /* key = XDeviceInfo, value = GsdWacomDevice */

	/* button capture */
	GSList *screens;
	int      opcode;
};

static GdkFilterReturn filter_button_events   (XEvent          *xevent,
                                               GdkEvent        *event,
                                               GsdWacomManager *manager);

static void     gsd_wacom_manager_class_init  (GsdWacomManagerClass *klass);
static void     gsd_wacom_manager_init	(GsdWacomManager      *wacom_manager);
static void     gsd_wacom_manager_finalize    (GObject	      *object);

G_DEFINE_TYPE (GsdWacomManager, gsd_wacom_manager, G_TYPE_OBJECT)

static gpointer manager_object = NULL;

static GObject *
gsd_wacom_manager_constructor (GType		     type,
			       guint		      n_construct_properties,
			       GObjectConstructParam     *construct_properties)
{
	GsdWacomManager      *wacom_manager;

	wacom_manager = GSD_WACOM_MANAGER (G_OBJECT_CLASS (gsd_wacom_manager_parent_class)->constructor (type,
													 n_construct_properties,
													 construct_properties));

	return G_OBJECT (wacom_manager);
}

static void
gsd_wacom_manager_dispose (GObject *object)
{
	G_OBJECT_CLASS (gsd_wacom_manager_parent_class)->dispose (object);
}

static void
gsd_wacom_manager_class_init (GsdWacomManagerClass *klass)
{
	GObjectClass   *object_class = G_OBJECT_CLASS (klass);

	object_class->constructor = gsd_wacom_manager_constructor;
	object_class->dispose = gsd_wacom_manager_dispose;
	object_class->finalize = gsd_wacom_manager_finalize;

	g_type_class_add_private (klass, sizeof (GsdWacomManagerPrivate));
}

static int
get_device_id (GsdWacomDevice *device)
{
	XDeviceInfo *device_info;
	int id;

	g_object_get (device, "device-info", &device_info, NULL);
	if (device_info == NULL)
		return -1;
	id = device_info->id;
	return id;
}

static XDevice *
open_device (GsdWacomDevice *device)
{
	XDevice *xdev;
	int id;

	id = get_device_id (device);
	if (id < 0)
		return NULL;

	gdk_error_trap_push ();
	xdev = XOpenDevice (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), id);
	if (gdk_error_trap_pop () || (xdev == NULL))
		return NULL;

	return xdev;
}

static void
wacom_set_property (GsdWacomDevice *device,
		    PropertyHelper *property)
{
	XDevice *xdev;

	xdev = open_device (device);
	device_set_property (xdev, gsd_wacom_device_get_tool_name (device), property);
	XCloseDevice (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdev);
}

static void
set_rotation (GsdWacomDevice *device,
	      GsdWacomRotation rotation)
{
	gchar rot = rotation;
	PropertyHelper property = {
		.name = "Wacom Rotation",
		.nitems = 1,
		.format = 8,
		.type   = XA_INTEGER,
		.data.c = &rot,
	};

	wacom_set_property (device, &property);
}

static void
set_pressurecurve (GsdWacomDevice *device,
		   GConfValue     *value)
{
	PropertyHelper property = {
		.name = "Wacom Pressurecurve",
		.nitems = 4,
		.type   = XA_INTEGER,
		.format = 32,
	};
	gsize nvalues;
	int *int_array;

	int_array = gsd_gconf_value_get_int_array (value, &nvalues);
	if (nvalues != 4) {
		g_free (int_array);
		g_error ("Pressurecurve requires 4 values.");
		return;
	}

	property.data.i = int_array;
	wacom_set_property (device, &property);
	g_free(int_array);
}

/* Area handling. Each area is defined as top x/y, bottom x/y and limits the
 * usable area of the physical device to the given area (in device coords)
 */
static void
set_area (GsdWacomDevice  *device,
	  GConfValue      *value)
{
	PropertyHelper property = {
		.name = "Wacom Tablet Area",
		.nitems = 4,
		.type   = XA_INTEGER,
		.format = 32,
	};
	gsize nvalues;
	int *int_array;

	int_array = gsd_gconf_value_get_int_array (value, &nvalues);
	if (nvalues != 4) {
		g_error ("Area configuration requires 4 values, got %i.", (gint) nvalues);
		return;
	}

	property.data.i = int_array;
	wacom_set_property (device, &property);
	g_free(int_array);
}

static void
set_display (GsdWacomDevice  *device,
	     GConfValue      *value)
{
	float matrix[NUM_ELEMS_MATRIX];
	PropertyHelper property = {
		.name   = "Coordinate Transformation Matrix",
		.nitems = NUM_ELEMS_MATRIX,
		.format = 32,
		.type   = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), "FLOAT", True),
	};

	gsd_wacom_device_get_display_matrix (device, matrix);

	property.data.i = (gint*)(&matrix);
	g_debug ("Applying matrix to device...");
	wacom_set_property (device, &property);
}

static void
set_absolute (GsdWacomDevice  *device,
	      gint	     is_absolute)
{
	XDevice *xdev;

	xdev = open_device (device);
	gdk_error_trap_push ();
	XSetDeviceMode (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdev, is_absolute ? Absolute : Relative);
	if (gdk_error_trap_pop ())
		g_error ("Failed to set mode \"%s\" for \"%s\".",
			 is_absolute ? "Absolute" : "Relative", gsd_wacom_device_get_tool_name (device));
	XCloseDevice (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdev);
}

static void
set_device_buttonmap (GsdWacomDevice *device,
		      GConfValue      *value)
{
	XDevice *xdev;
	gsize nmap;
	gint *intmap;
	unsigned char *map;
	int i, j, rc;

	xdev = open_device (device);

	intmap = gsd_gconf_value_get_int_array (value, &nmap);
	map = g_new0 (unsigned char, nmap);
	for (i = 0; i < nmap; i++)
		map[i] = intmap[i];
	g_free(intmap);

	gdk_error_trap_push ();

	/* X refuses to change the mapping while buttons are engaged,
	 * so if this is the case we'll retry a few times
	 */
	for (j = 0;
	     j < 20 && (rc = XSetDeviceButtonMapping (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdev, map, nmap)) == MappingBusy;
	     ++j) {
		g_usleep (300);
	}

	if ((gdk_error_trap_pop () && rc != MappingSuccess) ||
	    rc != MappingSuccess)
		g_warning ("Error in setting button mapping for \"%s\"", gsd_wacom_device_get_tool_name (device));

	g_free (map);

	XCloseDevice (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdev);
}

static void
set_touch (GsdWacomDevice *device,
	   gboolean	touch)
{
	gchar data = touch;
	PropertyHelper property = {
		.name = "Wacom Enable Touch",
		.nitems = 1,
		.format = 8,
		.type   = XA_INTEGER,
		.data.c = &data,
	};

	wacom_set_property (device, &property);
}

static void
set_tpcbutton (GsdWacomDevice *device,
	       gboolean	tpcbutton)
{
	/* Wacom's TPCButton option which this setting emulates is to enable
	 * Tablet PC stylus behaviour when on. The property "Hover Click"
	 * works the other way round, i.e. if Hover Click is enabled this
	 * is the equivalent of TPC behaviour disabled. */
	gchar data = tpcbutton ? 0 : 1;
	PropertyHelper property = {
		.name = "Wacom Hover Click",
		.nitems = 1,
		.format = 8,
		.type   = XA_INTEGER,
		.data.c = &data,
	};

	wacom_set_property (device, &property);
}

static void
set_pressurethreshold (GsdWacomDevice *device,
		       gint	    threshold)
{
	PropertyHelper property = {
		.name = "Wacom Pressure Threshold",
		.nitems = 1,
		.format = 32,
		.type   = XA_INTEGER,
		.data.i = &threshold,
	};

	wacom_set_property (device, &property);
}

static void
apply_stylus_settings (GsdWacomDevice *device)
{
	GConfValue *value;
	GsdWacomStylus *stylus;
	int threshold;

	g_object_get (device, "last-stylus", &stylus, NULL);
	if (stylus == NULL) {
		g_warning ("Last stylus is not set");
		return;
	}

	g_debug ("Applying setting for stylus '%s' on device '%s'",
		 gsd_wacom_stylus_get_name (stylus),
		 gsd_wacom_device_get_name (device));

	value = gsd_gconf_get (gsd_wacom_stylus_get_settings_path (stylus),
			       gsd_wacom_stylus_get_schemas_path (stylus),
			       KEY_PRESSURECURVE);
	set_pressurecurve (device, value);
	gconf_value_free (value);

	value = gsd_gconf_get (gsd_wacom_stylus_get_settings_path (stylus),
			       gsd_wacom_stylus_get_schemas_path (stylus),
			       KEY_BUTTON_MAPPING);
	set_device_buttonmap (device, value);
	gconf_value_free (value);

	value = gsd_gconf_get (gsd_wacom_stylus_get_settings_path (stylus),
			       gsd_wacom_stylus_get_schemas_path (stylus),
			       KEY_PRESSURETHRESHOLD);
	threshold = gconf_value_get_int (value);
	if (threshold == -1)
		threshold = DEFAULT_PRESSURE_THRESHOLD;
	set_pressurethreshold (device, threshold);
	gconf_value_free (value);
}

/*
 * The rule to determine the status LED to use is as follow:
 *
 * "[...] if a device has only one ring/strip, use status_led0_select;
 *  otherwise the left ring/strip is controlled by status_led1_select and
 *  the right ring/strip by status_led0_select."
 *
 * http://sourceforge.net/mailarchive/message.php?msg_id=29898591
 */
static int
get_led_group_id(GsdWacomDevice *device,
		 int             group_id)
{
	gint num_rings;
	gint num_strips;

	num_rings = gsd_wacom_device_get_num_rings (device);
	num_strips = gsd_wacom_device_get_num_strips (device);

	/* Given group_id is in {1..4} as follow
	 * WACOM_BUTTON_RING_MODESWITCH        => group_id == 1
	 * WACOM_BUTTON_RING2_MODESWITCH       => group_id == 2
	 * WACOM_BUTTON_TOUCHSTRIP_MODESWITCH  => group_id == 3
	 * WACOM_BUTTON_TOUCHSTRIP2_MODESWITCH => group_id == 4
	 *
	 * see function flags_to_group() in gsd-wacom-device.c
	 */

	if ((num_rings == 1) && (group_id == 1))
		return 0;

	if ((num_strips == 1) && (group_id == 3))
		return 0;

	if ((num_rings == 2) && (group_id == 1 ||  group_id == 2))
		return (group_id & 1);

	if ((num_strips == 2) && (group_id == 3 ||  group_id == 4))
		return (group_id & 1);

	g_debug ("Group ID %d not applicable for %d ring(s) and %d strip(s)",
		 group_id, num_rings, num_strips);

	return -1;
}

static void
set_led (GsdWacomDevice *device,
	 int             group_id,
	 int             index)
{
	GError *error = NULL;
	const char *path;
	char *command;
	gint status_led;
	gboolean ret;

#ifndef HAVE_GUDEV
	/* Not implemented on non-Linux systems */
	return;
#endif
	g_return_if_fail (index >= 1);

	path = gsd_wacom_device_get_path (device);
	status_led = get_led_group_id (device, group_id);

	if (status_led < 0) {
		g_debug ("Ignoring group ID %d for device %s",
		         group_id, gsd_wacom_device_get_name (device));
		return;
	}

	g_debug ("Switching group ID %d to index %d for device %s", group_id, index, path);

	command = g_strdup_printf ("pkexec " LIBEXECDIR "/gsd-wacom-led-helper --path %s --group %d --led %d",
				   path, status_led, index - 1);
	ret = g_spawn_command_line_sync (command,
					 NULL,
					 NULL,
					 NULL,
					 &error);

	if (ret == FALSE) {
		g_debug ("Failed to launch '%s': %s", command, error->message);
		g_error_free (error);
	}

	g_free (command);
}

struct DefaultButtons {
	const char *button;
	int	 num;
};

struct DefaultButtons def_touchrings_buttons[] = {
	/* Touchrings */
	{ "AbsWheelUp", 90 },
	{ "AbsWheelDown", 91 },
	{ "RelWheelUp", 90 },
	{ "RelWheelDown", 91 },
	{ "AbsWheel2Up", 92 },
	{ "AbsWheel2Down", 93 },
	{ NULL, 0 }
};

struct DefaultButtons def_touchstrip_buttons[] = {
	/* Touchstrips */
	{ "StripLeftUp", 94 },
	{ "StripLeftDown", 95 },
	{ "StripRightUp", 96 },
	{ "StripRightDown", 97 },
	{ NULL, 0 }
};

static void
reset_touch_buttons (XDevice               *xdev,
		     struct DefaultButtons *buttons,
		     const char            *device_property)
{
	Atom actions[6];
	Atom action_prop;
	guint i;

	/* Create a device property with the action for button i */
	for (i = 0; buttons[i].button != NULL; i++)
	{
		char *propname;
		int action[2]; /* press + release */
		Atom prop;
		int mapped_button = buttons[i].num;

		action[0] = AC_BUTTON | AC_KEYBTNPRESS | mapped_button;
		action[1] = AC_BUTTON | mapped_button;

		propname = g_strdup_printf ("Button %s action", buttons[i].button);
		prop = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), propname, False);
		g_free (propname);
		XChangeDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdev,
				       prop, XA_INTEGER, 32, PropModeReplace,
				       (const guchar *) &action, 2);

		/* prop now contains press + release for the mapped button */
		actions[i] = prop;
	}

	/* Now set the actual action property to contain references to the various
	 * actions */
	action_prop = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), device_property, True);
	XChangeDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdev,
			       action_prop, XA_ATOM, 32, PropModeReplace,
			       (const guchar *) actions, i);
}

static void
reset_pad_buttons (GsdWacomDevice *device)
{
	XDevice *xdev;
	int nmap;
	unsigned char *map;
	int i, j, rc;

	/* Normal buttons */
	xdev = open_device (device);

	gdk_error_trap_push ();

	nmap = 256;
	map = g_new0 (unsigned char, nmap);
	for (i = 0; i < nmap && i < sizeof (map); i++)
		map[i] = i + 1;

	/* X refuses to change the mapping while buttons are engaged,
	 * so if this is the case we'll retry a few times */
	for (j = 0;
	     j < 20 && (rc = XSetDeviceButtonMapping (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdev, map, nmap)) == MappingBusy;
	     ++j) {
		g_usleep (300);
	}

	if ((gdk_error_trap_pop () && rc != MappingSuccess) ||
	    rc != MappingSuccess)
		g_warning ("Error in resetting button mapping for \"%s\" (rc=%d)", gsd_wacom_device_get_tool_name (device), rc);

	g_free (map);

	gdk_error_trap_push ();
	reset_touch_buttons (xdev, def_touchrings_buttons, "Wacom Wheel Buttons");
	reset_touch_buttons (xdev, def_touchstrip_buttons, "Wacom Strip Buttons");
	gdk_error_trap_pop ();

	XCloseDevice (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdev);

	/* Reset all the LEDs */
	/* FIXME, get the number of modes somewhere else */
	for (i = 1; i <= 4; i++)
		set_led (device, i, 1);
}

static void
set_wacom_settings (GsdWacomManager *manager,
		    GsdWacomDevice  *device)
{
	GsdWacomDeviceType type;
	GConfValue *value;

	g_debug ("Applying settings for device '%s' (type: %s)",
		 gsd_wacom_device_get_tool_name (device),
		 gsd_wacom_device_type_to_string (gsd_wacom_device_get_device_type (device)));

	value = gsd_gconf_get (gsd_wacom_device_get_settings_path (device),
			       gsd_wacom_device_get_schemas_path (device),
			       KEY_ROTATION);
	set_rotation (device, gsd_wacom_get_rotation_type_from_name (gconf_value_get_string (value)));
	gconf_value_free (value);

	value = gsd_gconf_get (gsd_wacom_device_get_settings_path (device),
			       gsd_wacom_device_get_schemas_path (device),
			       KEY_TOUCH);
	set_touch (device, gconf_value_get_bool (value));
	gconf_value_free (value);

	type = gsd_wacom_device_get_device_type (device);

	if (type == WACOM_TYPE_TOUCH &&
	    gsd_wacom_device_is_screen_tablet (device) == FALSE) {
		set_absolute (device, FALSE);
		return;
	}

	if (type == WACOM_TYPE_CURSOR) {
		int values[] = {-1, -1, -1, -1};

		set_absolute (device, FALSE);

		value = gconf_value_new (GCONF_VALUE_LIST);
		gconf_value_set_list_type (value, GCONF_VALUE_INT);
		gsd_gconf_value_set_int_array (value, values, G_N_ELEMENTS (values));

		set_area (device, value);
		gconf_value_free (value);

		return;
	}

	if (type == WACOM_TYPE_PAD) {
		int id;

		id = get_device_id (device);
		reset_pad_buttons (device);
		grab_button (id, TRUE, manager->priv->screens);
		return;
	}

	if (type == WACOM_TYPE_STYLUS) {
		value = gsd_gconf_get (gsd_wacom_device_get_settings_path (device),
				       gsd_wacom_device_get_schemas_path (device),
				       KEY_TPCBUTTON);
		set_tpcbutton (device, gconf_value_get_bool (value));
		gconf_value_free (value);
	}

	value = gsd_gconf_get (gsd_wacom_device_get_settings_path (device),
			       gsd_wacom_device_get_schemas_path (device),
			       KEY_IS_ABSOLUTE);
	set_absolute (device, gconf_value_get_bool (value));
	gconf_value_free (value);

	/* Ignore touch devices as they do not share the same range of values for area */
	if (type != WACOM_TYPE_TOUCH) {
		value = gsd_gconf_get (gsd_wacom_device_get_settings_path (device),
				       gsd_wacom_device_get_schemas_path (device),
				       KEY_AREA);
		set_area (device, value);
		gconf_value_free (value);
	}

	value = gsd_gconf_get (gsd_wacom_device_get_settings_path (device),
			       gsd_wacom_device_get_schemas_path (device),
			       KEY_DISPLAY);
	set_display (device, value);
	gconf_value_free (value);

	/* only pen and eraser have pressure threshold and curve settings */
	if (type == WACOM_TYPE_STYLUS ||
	    type == WACOM_TYPE_ERASER) {
		apply_stylus_settings (device);
	}
}

static void
wacom_settings_changed (GConfValue     *value,
			gchar          *key,
			GsdWacomDevice *device)
{
	GsdWacomDeviceType type;

	type = gsd_wacom_device_get_device_type (device);

	if (g_str_equal (key, KEY_ROTATION)) {
	        if (type != WACOM_TYPE_PAD)
		        set_rotation (device, gsd_wacom_get_rotation_type_from_name (gconf_value_get_string (value)));
	} else if (g_str_equal (key, KEY_TOUCH)) {
		set_touch (device, gconf_value_get_bool (value));
	} else if (g_str_equal (key, KEY_TPCBUTTON)) {
		set_tpcbutton (device, gconf_value_get_bool (value));
	} else if (g_str_equal (key, KEY_IS_ABSOLUTE)) {
		if (type != WACOM_TYPE_CURSOR &&
		    type != WACOM_TYPE_PAD &&
		    type != WACOM_TYPE_TOUCH)
			set_absolute (device, gconf_value_get_bool (value));
	} else if (g_str_equal (key, KEY_AREA)) {
		if (type != WACOM_TYPE_CURSOR &&
		    type != WACOM_TYPE_PAD &&
		    type != WACOM_TYPE_TOUCH)
			set_area (device, value);
	} else if (g_str_equal (key, KEY_DISPLAY)) {
		if (type != WACOM_TYPE_CURSOR &&
		    type != WACOM_TYPE_PAD)
			set_display (device, value);
	} else {
		g_warning ("Unhandled tablet-wide setting '%s' changed", key);
	}
}

static void
stylus_settings_changed (GConfValue     *value,
			 gchar          *key,
			 GsdWacomStylus *stylus)
{
	GsdWacomDevice *device;
	GsdWacomStylus *last_stylus;

	device = gsd_wacom_stylus_get_device (stylus);

	g_object_get (device, "last-stylus", &last_stylus, NULL);
	if (last_stylus != stylus) {
		g_debug ("Not applying changed settings because '%s' is the current stylus, not '%s'",
			 last_stylus ? gsd_wacom_stylus_get_name (last_stylus) : "NONE",
			 gsd_wacom_stylus_get_name (stylus));
		return;
	}

	if (g_str_equal (key, KEY_PRESSURECURVE)) {
		set_pressurecurve (device, value);
	} else if (g_str_equal (key, KEY_PRESSURETHRESHOLD)) {
		int threshold;

		threshold = gconf_value_get_int (value);
		if (threshold == -1)
			threshold = DEFAULT_PRESSURE_THRESHOLD;
		set_pressurethreshold (device, threshold);
	} else if (g_str_equal (key, KEY_BUTTON_MAPPING)) {
		set_device_buttonmap (device, value);
	} else {
		g_warning ("Unhandled stylus setting '%s' changed", key);
	}
}

static void
last_stylus_changed (GsdWacomDevice  *device)
{
	g_debug ("Stylus for device '%s' changed, applying settings",
		 gsd_wacom_device_get_name (device));
	apply_stylus_settings (device);
}

static gboolean
setup_event_mask (GdkScreen *gdk_screen, int device_id)
{
	unsigned char mask[1] = { 0 };
	Display *dpy;
	XIEventMask evmask;

	g_return_val_if_fail (GDK_IS_SCREEN (gdk_screen), FALSE);

	evmask.deviceid = device_id;
	evmask.mask_len = sizeof(mask);
	evmask.mask = mask;
	XISetMask (evmask.mask, XI_ButtonPress);
	XISetMask (evmask.mask, XI_ButtonRelease);

	gdk_error_trap_push ();
	dpy = GDK_DISPLAY_XDISPLAY (gdk_screen_get_display (gdk_screen));
	XISelectEvents (dpy, DefaultRootWindow(dpy), &evmask, 1);
	if (gdk_error_trap_pop ())
		return FALSE;

	return TRUE;
}

static void
device_added_cb (GsdDeviceManager *device_manager,
		 XDeviceInfo      *device_info,
		 GsdWacomManager  *manager)
{
	GsdWacomDevice *device;
	GConfClient    *client;
	const char     *settings_path;
	GSList *ls;

	device = gsd_wacom_device_new (device_info);
	if (gsd_wacom_device_get_device_type (device) == WACOM_TYPE_INVALID) {
		g_debug ("Device '%s' is invalid", device_info->name);
		g_object_unref (device);
		return;
	}
	g_debug ("Adding device '%s' (type: '%s') to known devices list",
		 gsd_wacom_device_get_tool_name (device),
		 gsd_wacom_device_type_to_string (gsd_wacom_device_get_device_type (device)));
	g_hash_table_insert (manager->priv->devices, (gpointer) device_info, device);

	if (gsd_wacom_device_get_device_type (device) == WACOM_TYPE_STYLUS ||
	    gsd_wacom_device_get_device_type (device) == WACOM_TYPE_ERASER) {
		g_signal_connect (G_OBJECT (device), "notify::last-stylus",
				  G_CALLBACK (last_stylus_changed), manager);
	}
	set_wacom_settings (manager, device);

	/* Set event mask for this device id */
	for (ls = manager->priv->screens; ls != NULL; ls = ls->next)
		setup_event_mask (ls->data, device_info->id);

	/* GConf backport */
	settings_path = gsd_wacom_device_get_settings_path (device);
	if (settings_path) {
		client = gconf_client_get_default ();
		gconf_client_add_dir (client, settings_path, GCONF_CLIENT_PRELOAD_NONE, NULL);
		g_object_unref (client);
	}
}

static void
device_removed_cb (GsdDeviceManager *device_manager,
		   XDeviceInfo      *device_info,
		   GsdWacomManager  *manager)
{
	GsdWacomDevice *device;
	GConfClient    *client;
	const char     *settings_path;

	g_debug ("Removing device '%s' from known devices list",
		 device_info->name);
	device = g_hash_table_lookup (manager->priv->devices, device_info);
	if (device) {
		settings_path = gsd_wacom_device_get_settings_path (device);
		if (settings_path) {
			client = gconf_client_get_default ();
			gconf_client_remove_dir (client, settings_path, NULL);
			g_object_unref (client);
		}
	}
	g_hash_table_remove (manager->priv->devices, device_info);

	/* Enable this chunk of code if you want to valgrind
	 * test-wacom. It will exit when there are no Wacom devices left */
#if 0
	if (g_hash_table_size (manager->priv->devices) == 0)
		gtk_main_quit ();
#endif
}

static GsdWacomDevice *
device_id_to_device (GsdWacomManager *manager,
		     int              deviceid)
{
	GList *devices, *l;
	GsdWacomDevice *ret;

	ret = NULL;
	devices = g_hash_table_get_keys (manager->priv->devices);

	for (l = devices; l != NULL; l = l->next) {
		XDeviceInfo *device_info = l->data;
		int id;

		id = device_info->id;
		if (id == deviceid) {
			ret = g_hash_table_lookup (manager->priv->devices, device_info);
			break;
		}
	}

	g_list_free (devices);
	return ret;
}

struct {
	guint mask;
	KeySym keysym;
} mods_keysyms[] = {
	{ GDK_MOD1_MASK, XK_Alt_L },
	{ GDK_SHIFT_MASK, XK_Shift_L },
	{ GDK_CONTROL_MASK, XK_Control_L },
};

static void
send_modifiers (Display *display,
		guint mask,
		gboolean is_press)
{
	guint i;

	if (mask == 0)
		return;

	for (i = 0; i < G_N_ELEMENTS(mods_keysyms); i++) {
		if (mask & mods_keysyms[i].mask) {
			guint keycode;

			keycode = XKeysymToKeycode (display, mods_keysyms[i].keysym);
			XTestFakeKeyEvent (display, keycode,
					   is_press ? True : False, 0);
		}
	}
}

static char *
get_elevator_shortcut_string (GsdWacomTabletButton *wbutton,
			      GtkDirectionType      dir)
{
	char **strv, *str;
	GConfValue  *value;

	value = gsd_gconf_get (wbutton->settings_path, wbutton->schemas_path, KEY_CUSTOM_ELEVATOR_ACTION);
	strv = gsd_gconf_value_get_strv (value, NULL);
	gconf_value_free (value);
	if (strv == NULL)
		return NULL;

	if (g_strv_length (strv) >= 1 && dir == GTK_DIR_UP)
		str = g_strdup (strv[0]);
	else if (g_strv_length (strv) >= 2 && dir == GTK_DIR_DOWN)
		str = g_strdup (strv[1]);
	else
		str = NULL;

	g_strfreev (strv);

	return str;
}

static void
generate_key (GsdWacomTabletButton *wbutton,
	      int                   group,
	      Display              *display,
	      GtkDirectionType      dir,
	      gboolean              is_press)
{
	char                 *str;
	guint                 keyval;
	guint                *keycodes;
	guint                 keycode;
	guint                 mods;
	GdkKeymapKey         *keys;
	int                   n_keys;
	guint                 i;
	GConfValue           *value;

	if (wbutton->type == WACOM_TABLET_BUTTON_TYPE_ELEVATOR)
		str = get_elevator_shortcut_string (wbutton, dir);
	else {
		value = gsd_gconf_get (wbutton->settings_path, wbutton->schemas_path, KEY_CUSTOM_ACTION);
		str = g_strdup (gconf_value_get_string (value));
		gconf_value_free (value);
	}

	if (str == NULL)
		return;

	gsd_accelerator_parse_with_keycode (str, &keyval, &keycodes, &mods);
	if (keycodes == NULL) {
		g_warning ("Failed to find a keycode for shortcut '%s'", str);
		g_free (str);
		return;
	}
	g_free (keycodes);

	/* Now look for our own keycode, in the group as us */
	if (!gdk_keymap_get_entries_for_keyval (gdk_keymap_get_default (), keyval, &keys, &n_keys)) {
		g_warning ("Failed to find a keycode for keyval '%s' (0x%x)", gdk_keyval_name (keyval), keyval);
		g_free (str);
		return;
	}

	keycode = 0;
	for (i = 0; i < n_keys; i++) {
		if (keys[i].group != group)
			continue;
		if (keys[i].level > 0)
			continue;
		keycode = keys[i].keycode;
	}
	/* Couldn't find it in the current group? Look in group 0 */
	if (keycode == 0) {
		for (i = 0; i < n_keys; i++) {
			if (keys[i].group > 0)
				continue;
			keycode = keys[i].keycode;
		}
	}
	g_free (keys);

	if (keycode == 0) {
		g_warning ("Not emitting '%s' (keyval: %d, keycode: %d mods: 0x%x), invalid keycode",
			   str, keyval, keycode, mods);
		g_free (str);
		return;
	} else {
		g_debug ("Emitting '%s' (keyval: %d, keycode: %d mods: 0x%x)",
			 str, keyval, keycode, mods);
	}

	/* And send out the keys! */
	gdk_error_trap_push ();
	if (is_press)
		send_modifiers (display, mods, TRUE);
	XTestFakeKeyEvent (display, keycode,
			   is_press ? True : False, 0);
	if (is_press == FALSE)
		send_modifiers (display, mods, FALSE);
	if (gdk_error_trap_pop ())
		g_warning ("Failed to generate fake key event '%s'", str);

	g_free (str);
}

static GdkFilterReturn
filter_button_events (XEvent          *xevent,
                      GdkEvent        *event,
                      GsdWacomManager *manager)
{
	Display             *dpy;
	XIDeviceEvent       *xev;
	XGenericEventCookie *cookie;
	guint                deviceid;
	GsdWacomDevice      *device;
	int                  button;
	GsdWacomTabletButton *wbutton;
	GtkDirectionType      dir;
	const char           *str;
	GConfValue           *value;

	/* verify we have a key event */
	if (xevent->type != GenericEvent)
		return GDK_FILTER_CONTINUE;
	cookie = &xevent->xcookie;
	if (cookie->extension != manager->priv->opcode)
		return GDK_FILTER_CONTINUE;

	if (cookie->evtype != XI_ButtonRelease &&
	    cookie->evtype != XI_ButtonPress)
		return GDK_FILTER_CONTINUE;

	dpy = GDK_DISPLAY_XDISPLAY (gdk_display_get_default ());
	if (!XGetEventData(dpy, cookie))
		return GDK_FILTER_CONTINUE;

	xev = (XIDeviceEvent *) cookie->data;

	deviceid = xev->sourceid;
	device = device_id_to_device (manager, deviceid);
	if (gsd_wacom_device_get_device_type (device) != WACOM_TYPE_PAD)
		goto out;

	button = xev->detail;

	wbutton = gsd_wacom_device_get_button (device, button, &dir);
	if (wbutton == NULL) {
		g_warning ("Could not find matching button for '%d' on '%s'",
			   button, gsd_wacom_device_get_name (device));
		goto out;
	}

	g_debug ("Received event button %s '%s'%s ('%d') on device '%s' ('%d')",
		 xev->evtype == XI_ButtonPress ? "press" : "release",
		 wbutton->id,
		 wbutton->type == WACOM_TABLET_BUTTON_TYPE_ELEVATOR ?
		 (dir == GTK_DIR_UP ? " 'up'" : " 'down'") : "",
		 button,
		 gsd_wacom_device_get_name (device),
		 deviceid);

	if (wbutton->type == WACOM_TABLET_BUTTON_TYPE_HARDCODED) {
		int new_mode;

		/* We switch modes on key release */
		if (xev->evtype == XI_ButtonRelease)
			goto out;

		new_mode = gsd_wacom_device_set_next_mode (device, wbutton);
		set_led (device, wbutton->group_id, new_mode);
		goto out;
	}

	/* Nothing to do */
	value = gsd_gconf_get (wbutton->settings_path, wbutton->schemas_path, KEY_ACTION_TYPE);
	str = gconf_value_get_string (value);
	if (gsd_wacom_get_action_type_from_name (str) == GSD_WACOM_ACTION_TYPE_NONE) {
		gconf_value_free (value);
		goto out;
	}
	gconf_value_free (value);

	/* Send a key combination out */
	/* FIXME: we cannot use xev->display (#804907) */
	generate_key (wbutton, xev->group.effective, dpy, dir, xev->evtype == XI_ButtonPress ? True : False);

out:
	XFreeEventData(dpy, cookie);
	return GDK_FILTER_REMOVE;
}

static void
set_devicepresence_handler (GsdWacomManager *manager)
{
	GsdDeviceManager *device_manager;

	device_manager = gsd_display_get_device_manager (gdk_display_get_default ());
	if (device_manager == NULL)
		return;

	manager->priv->device_added_id = g_signal_connect (G_OBJECT (device_manager), "device-added",
							   G_CALLBACK (device_added_cb), manager);
	manager->priv->device_removed_id = g_signal_connect (G_OBJECT (device_manager), "device-removed",
							     G_CALLBACK (device_removed_cb), manager);
	manager->priv->device_manager = device_manager;
}

static GList *
get_device_list_for_path (GsdWacomManager    *manager,
			  const char	 *settings_path)
{
	GList *devices, *l, *matches;

	g_return_val_if_fail (GSD_IS_WACOM_MANAGER (manager), NULL);
	g_return_val_if_fail (settings_path != NULL, NULL);

	matches = NULL;
	devices = g_hash_table_get_values (manager->priv->devices);
	for (l = devices; l != NULL; l = l->next) {
		GsdWacomDevice *device = l->data;
		if (strcmp (gsd_wacom_device_get_settings_path (device), settings_path) == 0)
			matches = g_list_prepend (matches, device);
	}
	g_list_free (devices);

	return matches;
}


static gboolean
parse_entry (const char *entry, char **device, char **stylus, char **key)
{
	char *p;

	*device = *stylus = *key = NULL;

	p = g_strrstr (entry, "/");
	if (!p) {
		g_warning ("Cannot determine key from '%s'", entry);
		return FALSE;
	}
	*key = g_strdup (p + 1);
	*p = '\0';

	p = g_strrstr (entry, "/");
	if (!p) {
		*device = g_strdup (entry);
		return TRUE;
	}
	*stylus = g_strdup (p + 1);
	*p = '\0';
	*device = g_strdup (entry);

	return TRUE;
}
static void
wacom_callback (GConfClient	 *client,
		guint		  cnxn_id,
		GConfEntry	 *entry,
		GsdWacomManager  *manager)
{
	GList *wacom_device_list, *l;
	const char *prefix;
	char *key = NULL, *device_name = NULL, *stylus_name = NULL;
	char *entry_name = NULL, *device_path = NULL, *stylus_path = NULL;

	g_debug ("Notification '%s' changed", entry->key);

	prefix = g_strrstr (entry->key, GCONF_WACOM_DIR_SLASH);
	if (!prefix) {
		g_warning ("Ignoring unknown notification for entry '%s'", entry->key);
		return;
	}

	prefix += strlen (GCONF_WACOM_DIR_SLASH);
	entry_name = g_strdup (prefix); /* so we don't touch the GConfEntry data */
	if (!parse_entry (entry_name, &device_name, &stylus_name, &key)) {
		g_warning ("Failed to identify device and key for entry '%s'", entry_name);
		g_free (entry_name);
		return;
	}

	device_path = gsd_gconf_build_path (GCONF_WACOM_DIR, device_name);
	wacom_device_list = get_device_list_for_path (manager, device_path);

	for (l = wacom_device_list; l != NULL; l = l->next) {
		GsdWacomDevice *wacom_device = l->data;

		if (stylus_name) {
			GsdWacomStylus *wacom_stylus;

			stylus_path = gsd_gconf_build_path (device_path, stylus_name);
			wacom_stylus = gsd_wacom_device_get_stylus_for_path (wacom_device, stylus_path);
			g_free (stylus_path);

			if (wacom_stylus) {
				/* Apply setting to stylus */
				stylus_settings_changed (entry->value, key, wacom_stylus);
			}
		} else {
			/* Apply setting to device */
			g_debug ("Applying setting '%s' to device '%s'", entry->key, device_path);
			wacom_settings_changed (entry->value, key, wacom_device);
		}
	}

	g_list_free (wacom_device_list);
	g_free (device_path);
	g_free (entry_name);
	g_free (device_name);
	g_free (stylus_name);
	g_free (key);
}

static void
gsd_wacom_manager_init (GsdWacomManager *manager)
{
	manager->priv = GSD_WACOM_MANAGER_GET_PRIVATE (manager);
}

static gboolean
gsd_wacom_manager_idle_cb (GsdWacomManager *manager)
{
	GList *devices, *l;
	GConfClient *client;

	gnome_settings_profile_start (NULL);

	manager->priv->devices = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, g_object_unref);

	set_devicepresence_handler (manager);

	devices = gsd_device_manager_list_devices (manager->priv->device_manager, GSD_DEVICE_TYPE_SLAVE);
	for (l = devices; l ; l = l->next)
		device_added_cb (manager->priv->device_manager, l->data, manager);
	g_list_free (devices);

	client = gconf_client_get_default ();
	manager->priv->notify_wacom =
		gconf_client_notify_add (client,
					 GCONF_WACOM_DIR,
					 (GConfClientNotifyFunc) wacom_callback,
					 manager, NULL, NULL);
	g_object_unref (client);

	/* Start filtering the button events */
	gdk_window_add_filter (NULL,
			       (GdkFilterFunc) filter_button_events,
			       manager);

	gnome_settings_profile_end (NULL);

	manager->priv->start_idle_id = 0;

	return FALSE;
}

/*
 * The monitors-changed signal is emitted when the number, size or
 * position of the monitors attached to the screen change.
 */
static void
on_screen_changed_cb (GdkScreen *screen,
		      GsdWacomManager *manager)
{
	GList *devices, *l;

	/*
	 * A ::changed signal may be received at startup before
	 * the devices get added, in this case simply ignore the
	 * notification
	 */
	if (manager->priv->devices == NULL)
		return;

	g_debug ("Screen configuration changed");
	devices = g_hash_table_get_values (manager->priv->devices);
	for (l = devices; l != NULL; l = l->next) {
		GsdWacomDevice *device = l->data;
		GsdWacomDeviceType type;
		GConfValue *value;

		type = gsd_wacom_device_get_device_type (device);
		if (type == WACOM_TYPE_CURSOR || type == WACOM_TYPE_PAD)
			continue;

		value = gsd_gconf_get (gsd_wacom_device_get_settings_path (device),
				       gsd_wacom_device_get_schemas_path (device),
				       KEY_DISPLAY);
		set_display (device, value);
		gconf_value_free (value);
	}
	g_list_free (devices);
}

static void
init_screens (GsdWacomManager *manager)
{
        GdkDisplay *display;
        int i;

        display = gdk_display_get_default ();
        for (i = 0; i < gdk_display_get_n_screens (display); i++) {
                GError *error = NULL;
                GdkScreen *screen;

                screen = gdk_display_get_screen (display, i);
                if (screen == NULL) {
                        continue;
                }
                manager->priv->screens = g_slist_append (manager->priv->screens, screen);

		g_signal_connect (screen,
				  "monitors-changed",
				  G_CALLBACK (on_screen_changed_cb),
			          manager);

                /*
                 * When using GdkScreen, we need to update the input trans-
                 * formation matrix on overall ::size-changed signal as well
                 */
		g_signal_connect (screen,
				  "size-changed",
				  G_CALLBACK (on_screen_changed_cb),
			          manager);
        }
}

gboolean
gsd_wacom_manager_start (GsdWacomManager *manager,
			 GError	 **error)
{
	gnome_settings_profile_start (NULL);

	if (supports_xinput2_devices (&manager->priv->opcode) == FALSE) {
		g_debug ("No Xinput2 support, disabling plugin");
		return TRUE;
	}

	init_screens (manager);

	manager->priv->start_idle_id = g_idle_add ((GSourceFunc) gsd_wacom_manager_idle_cb, manager);

	gnome_settings_profile_end (NULL);

	return TRUE;
}

void
gsd_wacom_manager_stop (GsdWacomManager *manager)
{
	GsdWacomManagerPrivate *p = manager->priv;
	GSList *ls;

	g_debug ("Stopping wacom manager");

	if (p->device_manager != NULL) {
		g_signal_handler_disconnect (p->device_manager, p->device_added_id);
		g_signal_handler_disconnect (p->device_manager, p->device_removed_id);
		p->device_manager = NULL;
	}

	gdk_window_remove_filter (NULL,
				  (GdkFilterFunc) filter_button_events,
				  manager);

	for (ls = p->screens; ls != NULL; ls = ls->next)
		g_signal_handlers_disconnect_by_func (ls->data, on_screen_changed_cb, manager);
}

static void
gsd_wacom_manager_finalize (GObject *object)
{
	GsdWacomManager *wacom_manager;

	g_return_if_fail (object != NULL);
	g_return_if_fail (GSD_IS_WACOM_MANAGER (object));

	wacom_manager = GSD_WACOM_MANAGER (object);

	g_return_if_fail (wacom_manager->priv != NULL);

	if (wacom_manager->priv->devices) {
		g_hash_table_destroy (wacom_manager->priv->devices);
		wacom_manager->priv->devices = NULL;
	}

	if (wacom_manager->priv->screens != NULL) {
		g_slist_free (wacom_manager->priv->screens);
		wacom_manager->priv->screens = NULL;
	}

        if (wacom_manager->priv->start_idle_id != 0)
                g_source_remove (wacom_manager->priv->start_idle_id);

	G_OBJECT_CLASS (gsd_wacom_manager_parent_class)->finalize (object);
}

GsdWacomManager *
gsd_wacom_manager_new (void)
{
	if (manager_object != NULL) {
		g_object_ref (manager_object);
	} else {
		manager_object = g_object_new (GSD_TYPE_WACOM_MANAGER, NULL);
		g_object_add_weak_pointer (manager_object,
					   (gpointer *) &manager_object);
	}

	return GSD_WACOM_MANAGER (manager_object);
}
